"""manage all available computing resources needed in fireworks"""
import os
import sys
import grp
import getpass
import numbers
import uuid
import typing
import warnings
import json
from dataclasses import dataclass, field
from subprocess import PIPE, Popen
from semantic_version import Version, SimpleSpec
from fireworks.utilities.fw_utilities import explicit_serialize
from fireworks.utilities.fw_serializers import FWSerializable
from fireworks.utilities.fw_serializers import serialize_fw
from fireworks.utilities.fw_serializers import recursive_serialize
from fireworks.utilities.fw_serializers import recursive_deserialize
import fireworks_schema
from virtmat.middleware.exceptions import ResourceConfigurationError, SlurmError
from virtmat.middleware.utilities import lmod_env_module

SCHEMA = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'resconfig.json')
fireworks_schema.register_schema(SCHEMA)


def get_resconfig_loc():
    """return the resconfig path"""
    if 'RESCONFIG_LOC' in os.environ:
        return os.environ['RESCONFIG_LOC']
    return os.path.join(os.path.expanduser('~'), '.fireworks', 'res_config.yaml')


def get_default_resconfig():
    """return default resconfig"""
    resconfig_loc = get_resconfig_loc()
    if not os.path.exists(resconfig_loc):
        return None
    try:
        return ResConfig.from_file(resconfig_loc)
    except Exception as err:
        msg = f'Error occured loading configuration file {resconfig_loc}.'
        raise ResourceConfigurationError(msg) from err


def configure(resconfig_loc):
    """create default resconfig and resconfig_loc path, and write resconfig"""
    try:
        cfg = ResConfig.from_scratch()
        set_defaults_from_guess(cfg.default_worker)
    except SlurmError as err:
        print(err)
    except ResourceConfigurationError as err:
        print(err)
    else:
        resconfig_dir = os.path.dirname(resconfig_loc)
        if not os.path.exists(resconfig_dir):
            os.makedirs(resconfig_dir, exist_ok=True)
        cfg.to_file(resconfig_loc)


def get_env_modules(typ):
    """return a list of environment modules with their versions"""
    cmd_map = {'loaded': 'list', 'available': 'spider'}
    mods_dct = lmod_env_module(cmd_map[typ])
    mods_lst = []
    if mods_dct is None:
        return mods_lst
    for mod_name, versions in mods_dct.items():
        divisions = mod_name.split('/')
        prefix = '/'.join(divisions[:-1]) or None
        mods_lst.append(ModuleConfig(prefix, divisions[-1], versions))
    return mods_lst


def get_venv():
    """return a dictionary with venv type, name and path of activated"""
    if os.environ.get('CONDA_DEFAULT_ENV') is not None:
        return {'type': 'conda', 'name': os.environ.get('CONDA_DEFAULT_ENV'),
                'prefix': os.environ.get('CONDA_PREFIX')}
    if sys.prefix != sys.base_prefix:
        return {'type': 'venv', 'name': sys.prefix.rsplit('/', maxsplit=1)[-1],
                'prefix': sys.prefix}
    return None


class ResConfigSerializable(FWSerializable):
    """base serialization class for resource classes"""

    @serialize_fw
    @recursive_serialize
    def to_dict(self):
        return self.__dict__

    @classmethod
    @recursive_deserialize
    def from_dict(cls, m_dict):
        assert getattr(cls, '_fw_name') == m_dict.pop('_fw_name')
        return cls(**m_dict)


def resource_type_checker(func):
    """decorate the menimum, maximum and default methods of ResourceConfig"""
    def wrapper(*args, **kwargs):
        if (isinstance(args[1], bool) or
           not (isinstance(args[1], (int, float)) or args[1] is None)):
            raise ResourceConfigurationError('resource must be real numeric type')
        return func(*args, **kwargs)
    return wrapper


@dataclass
@explicit_serialize
class ResourceConfig(ResConfigSerializable):
    """Computing resource

    Attributes:
        name: the name of the resource
    """
    name: str
    _minimum: typing.Union[int, float] = None
    _maximum: typing.Union[int, float] = None
    _default: typing.Union[int, float] = None

    @property
    def minimum(self) -> typing.Union[int, float]:
        """The minimum value of the resource

        Returns:
            (float|int): the minimum value of the resource
        """
        return self._minimum

    @minimum.setter
    @resource_type_checker
    def minimum(self, val) -> None:
        if val is not None:
            if not isinstance(val, numbers.Number):
                msg = f'value must be numeric type: {val}'
                raise ResourceConfigurationError(msg)
            if val < 0:
                msg = f'negative value is not allowed: {val}'
                raise ResourceConfigurationError(msg)
            if self._maximum is not None and val > self._maximum:
                self._maximum = None
            if self._default is not None and val > self._default:
                self._default = None
        self._minimum = val

    @property
    def maximum(self):
        """The maximum value of the resource

        Returns:
            (float|int): the maximum value of the resource
        """
        return self._maximum

    @maximum.setter
    @resource_type_checker
    def maximum(self, val):
        if val is not None:
            if not isinstance(val, numbers.Number):
                msg = f'value must be numeric type: {val}'
                raise ResourceConfigurationError(msg)
            if val < 0:
                msg = f'negative value is not allowed: {val}'
                raise ResourceConfigurationError(msg)
            if self._minimum is not None and val < self._minimum:
                self._minimum = None
            if self._default is not None and val < self._default:
                self._default = None
        self._maximum = val

    @property
    def default(self):
        """The default value of the resource

        Returns:
            (float|int): the default value of the resource
        """
        return self._default

    @default.setter
    @resource_type_checker
    def default(self, val):
        if val is not None:
            if isinstance(val, numbers.Number) and val < 0:
                msg = f'negative value is not allowed: {val}'
                raise ResourceConfigurationError(msg)
            if self._maximum is not None and val > self._maximum:
                msg = f'cannot set default {val} greater than maximum {self._maximum}'
                raise ResourceConfigurationError(msg)
            if self._minimum is not None and val < self._minimum:
                msg = f'cannot set default {val} less than minimum {self._minimum}'
                raise ResourceConfigurationError(msg)
        self._default = val


@dataclass
@explicit_serialize
class QueueConfig(ResConfigSerializable):
    """Queue configuration

    Attributes:
        name (str): queue name
        public (bool): True if publicly available, default None
        resources ([ResourceConfig]): list of configured resources
        accounts_allow ([str]): list of usernames allowed to use the queue
        accounts_deny ([str]): list of usernames disallowed to use the queue
        groups_allow ([str]): list of groups allowed to use the queue
    """
    name: str = None
    public: bool = None
    resources: list[ResourceConfig] = field(default_factory=list)
    accounts_allow: list[str] = field(default_factory=list)
    accounts_deny: list[str] = field(default_factory=list)
    groups_allow: list[str] = field(default_factory=list)

    def set_resource(self, res, type_, val):
        """Set the value of a specific computing resource configuration

        Args:
            res (str): resource name
            type_ (str): resource type: default, minimum, maximum
            val (int|float): value of the resource

        Raises:
            ValueError: if type_ has invalid value
        """
        if type_ not in ('minimum', 'maximum', 'default'):
            raise ValueError(f'invalid resource type {type_}')
        if res in (r.name for r in self.resources):
            setattr(next(r for r in self.resources if res == r.name), type_, val)
        else:
            resource = ResourceConfig(name=res)
            setattr(resource, type_, val)
            self.resources.append(resource)

    def get_resource(self, res):
        """Get a specific computing resource configuration

        Args:
            res (str): resource name

        Returns:
            ResourceConfig object if resource exists, None otherwise
        """
        if res in (r.name for r in self.resources):
            return next(r for r in self.resources if r.name == res)
        return None

    def validate_resource(self, name, val):
        """Check if resource value is between minimum and maximum

        Args:
            name (str): resource name
            val (int|float): reource value

        Raises:
            ResourceConfigurationError: if resource value exceeds limits
        """
        if self.get_resource(name) is None:
            msg = f'resource {name} is not configured'
            raise ResourceConfigurationError(msg)
        if (self.get_resource(name).maximum is not None and
           val > self.get_resource(name).maximum):
            msg = f'resource {name} is greater than maximum'
            raise ResourceConfigurationError(msg)
        if (self.get_resource(name).minimum is not None and
           val < self.get_resource(name).minimum):
            msg = f'resource {name} is less than minimum'
            raise ResourceConfigurationError(msg)


@dataclass
@explicit_serialize
class ModuleConfig(ResConfigSerializable):
    """Module configuration

    Attributes:
        prefix (str): module prefix, if module name has prefix, else None
        name (str): module name, mandatory
        versions ([str]): available module versions, empty list if no versions
        path (str): module path, if different from default path, else None
    """
    prefix: str = None
    name: str = None
    versions: list[str] = field(default_factory=list)
    path: str = None

    def _format_command(self, version=None):
        """Format the module load command

        Args:
            version (str): optional version, no version if None

        Returns:
            (str): the formatted shell command to load the module
        """
        path_cmd = f'module use {self.path}; ' if self.path else ''
        comps = []
        if self.prefix:
            comps.append(self.prefix)
        comps.append(self.name)
        if version:
            assert version in self.versions
            comps.append(version)
        load_cmd = 'module load ' + '/'.join(comps)
        return path_cmd + load_cmd

    def get_command(self, name, spec=None):
        """Return the module load command

        Args:
            name (str): name to match
            spec (str): a requirement specification (like '<=0.9.3')

        Returns:
            (str): the shell command to load the module if matching else None
        """
        if name == self.name:
            if spec is None:
                if self.versions:
                    semvers = [Version.coerce(v) for v in self.versions]
                    ind = semvers.index(max(semvers))
                    return self._format_command(self.versions[ind])
                return self._format_command()
            spec_ = SimpleSpec(spec)
            for version in self.versions:
                if spec_.match(Version.coerce(version)):
                    return self._format_command(version)
        return None


@dataclass
@explicit_serialize
class CommandConfig(ResConfigSerializable):
    """Shell command configuration

    Attributes:
        cmd (str): a shell command
        args ([str]): a list of command arguments, default: empty list
        commands ([str]): a list of allowed commands

    Raises:
        ResourceConfigurationError: if a command is not allowed
    """
    cmd: str = None
    args: list[str] = field(default_factory=list)
    commands: list[str] = field(default_factory=lambda: ['umask', 'alias'])

    def __post_init__(self):
        if self.cmd not in self.commands:
            msg = f'command {self.cmd} not in the list of allowed commands'
            raise ResourceConfigurationError(msg)

    def get_command(self):
        """return the command as a string"""
        return self.cmd + ' ' + ' '.join(self.args)


@dataclass
@explicit_serialize
class WorkerConfig(ResConfigSerializable):
    """Worker configuration

    Every computing cluster is mapped to a worker.

    Attributes:
        name (str): worker name
        type_ (str): type of batch queuing system, only 'SLURM' supported
        queues ([QueueConfig]): a list of configured queues
        accounts ([str]): a list of groups available for accounting
        modules: ([ModuleConfig]): a list of available environment modules
        envvars: (dct): a dict of environment variables
        default_modules: ([ModuleConfig]): a list of default environment modules
        default_envvars: ([str]): a list of default environment variables
        default_venv: (dict): a dict of default virtual environment
        default_launchdir: (str): default directory for launching jobs
        default_commands: [CommandConfig]: a list of default commands
    """
    name: str = None
    type_: str = None
    queues: list[QueueConfig] = field(default_factory=list)
    accounts: list[str] = field(default_factory=list)
    modules: list[ModuleConfig] = field(default_factory=list)
    envvars: dict = field(default_factory=dict)
    default_modules: list[ModuleConfig] = field(default_factory=list)
    default_envvars: list[str] = field(default_factory=list)
    default_venv: dict = None
    default_launchdir: str = None
    default_commands: list[CommandConfig] = field(default_factory=list)
    _default_queue: QueueConfig = None
    _default_account: str = None

    @classmethod
    def from_scratch(cls, name=None):
        """Extract all relevant available resources from slurm partitions

        Args:
            name (str): optional worker name, if skipped one will be generated

        Returns:
            WorkerConfig object

        """
        name = uuid.uuid4().hex if name is None else name
        command = ['/usr/bin/scontrol', '--json', 'show', 'partition']
        try:
            with Popen(command, stdout=PIPE, shell=False) as proc:  # nosec B603
                slurm_res = json.load(proc.stdout)
        except FileNotFoundError as err:
            if 'No such file or directory: \'/usr/bin/scontrol\'' in str(err):
                return cls(name=name)
            raise ResourceConfigurationError('An unknown error occurred') from err
        queues = []
        for part in slurm_res['partitions']:
            q_kwargs = {}
            acc_allow = part['accounts']['allowed'].strip()
            acc_deny = part['accounts']['deny'].strip()
            grp_allow = part['groups']['allowed'].strip()
            q_kwargs['public'] = (acc_allow == '' and grp_allow == '') or acc_deny != ''
            if acc_allow:
                q_kwargs['accounts_allow'] = [a.strip() for a in acc_allow.split(',')]
            if acc_deny:
                q_kwargs['accounts_deny'] = [a.strip() for a in acc_deny.split(',')]
            if grp_allow:
                q_kwargs['groups_allow'] = [g.strip() for g in grp_allow.split(',')]
            q_kwargs['name'] = part['name']
            q_kwargs['resources'] = []
            queue = QueueConfig(**q_kwargs)
            for sec in ('maximums', 'minimums', 'defaults'):
                for key, val in part[sec].items():
                    if isinstance(val, dict):
                        if 'infinite' in val and 'number' in val:
                            value = float('inf') if val['infinite'] else val['number']
                        else:
                            continue
                    else:
                        value = val
                    try:
                        queue.set_resource(key, sec[:-1], value)
                    except ResourceConfigurationError:
                        queue.set_resource(key, sec[:-1], None)
                        msg = (f'resource {sec[:-1]} {key} has invalid value '
                               f'{val} or type {type(val)} and was set to null')
                        warnings.warn(msg, UserWarning)
            queues.append(queue)
        user = getpass.getuser()
        accounts = [g.gr_name for g in grp.getgrall() if user in g.gr_mem]
        return cls(name=name, type_='SLURM', queues=queues, accounts=accounts,
                   modules=get_env_modules('available'), default_venv=get_venv(),
                   default_modules=get_env_modules('loaded'))

    @property
    def default_account(self):
        """The default group used for accounting

        Returns:
            str: the default group used for accounting if set, otherwise None

        Raises:
            ResourceConfigurationError: if the group is not in the list of groups
        """
        return self._default_account

    @default_account.setter
    def default_account(self, acc=None):
        if acc not in self.accounts:
            msg = f'group {acc} not in the list of groups'
            raise ResourceConfigurationError(msg)
        self._default_account = acc

    def set_default_account(self):
        """select the currently active group as default group for accounting"""
        self.default_account = grp.getgrgid(os.getgid()).gr_name

    @property
    def default_queue(self) -> QueueConfig:
        """The default queue

        Returns:
            QueueConfig: the default queue configuration if set, otherwise None

        Raises:
            ResourceConfigurationError: if the queue is not in the list of queues
        """
        return self._default_queue

    @default_queue.setter
    def default_queue(self, queue: QueueConfig) -> None:
        if queue not in self.queues:
            msg = f'queue {queue} not in queues list'
            raise ResourceConfigurationError(msg)
        self._default_queue = queue

    def set_default_queue(self):
        """select the first configured queue as default queue"""
        if self._default_queue is None:
            if len(self.queues) == 0:
                raise ResourceConfigurationError('no queues configured')
            self._default_queue = self.queues[0]


@dataclass
@explicit_serialize
class ResConfig(ResConfigSerializable):
    """collect and store all computing resources needed to use fireworks"""
    workers: list[WorkerConfig] = field(default_factory=list)
    default_worker: WorkerConfig = None

    @classmethod
    def from_scratch(cls, name=None):
        """create a top-level resource configuration object from scratch"""
        worker = WorkerConfig.from_scratch(name)
        return cls(workers=[worker], default_worker=worker)

    def add_worker_from_scratch(self, name=None, default=False):
        """add a worker from scratch"""
        worker = WorkerConfig.from_scratch(name)
        self.workers.append(worker)
        if default:
            self.default_worker = worker

    @fireworks_schema.fw_schema_serialize
    def to_dict(self):
        return super().to_dict()


def set_defaults_from_guess(wcfg):
    """Set defaults of specific resources from a guess

    Args:
        wcfg (WorkerConfig): a worker configuration object
    """
    if wcfg.name is None:
        wcfg.name = uuid.uuid4().hex
    if len(wcfg.queues) > 0:
        if wcfg.default_queue is None:
            wcfg.set_default_queue()
        queue = wcfg.default_queue
        if wcfg.default_account is None:
            wcfg.set_default_account()
        time = queue.get_resource('time')
        if time is None:
            queue.set_resource('time', 'default', 5)
        else:
            if time.default is None:
                if time.minimum is None:
                    queue.set_resource('time', 'default', 5)
                else:
                    queue.set_resource('time', 'default', time.minimum)
        nodes = queue.get_resource('nodes')
        if nodes is None or nodes.default is None:
            queue.set_resource('nodes', 'default', 1)
        cpus_per_node = queue.get_resource('cpus_per_node')
        if cpus_per_node is None or cpus_per_node.default is None:
            queue.set_resource('cpus_per_node', 'default', 1)
        cpus_per_task = queue.get_resource('cpus_per_task')
        if cpus_per_task is None or cpus_per_task.default is None:
            queue.set_resource('cpus_per_task', 'default', 1)



"""export objects from resconfig, qadapter and cli modules"""
from .resconfig import ResourceConfig, QueueConfig, WorkerConfig, ResConfig
from .resconfig import ModuleConfig, CommandConfig
from .resconfig import get_resconfig_loc, get_default_resconfig
from .resconfig import set_defaults_from_guess, configure
from .qadapter import get_default_qadapter, get_custom_qadapter
from .qadapter import get_qadapter, get_worker_get_queue
from .cli import setup_resconfig



"""specific exceptions for the vre-middleware"""


class CustomException(Exception):
    """Custom exception class used for all specific exceptions"""


class ConfigurationException(CustomException):
    """Describes errors in the configuration of the function or missing parameters"""


class InvalidStateException(CustomException):
    """Describes errors while being in a wrong state"""
    def __init__(self, msg, node_id):
        super().__init__(f'Node ID: {str(node_id)}: {msg}')


class TestingException(CustomException):
    """Describes an exception while running a test"""
    __test__ = False  # this is to prevent pytest collecting the class


class SlurmError(RuntimeError):
    """Exception to raise if Slurm is not responding or not available"""


class ResourceConfigurationError(CustomException):
    """Exception to raise from the resconfig module"""



"""utility functions dealing with fireworks qadapter"""
import os
import math
from fireworks.user_objects.queue_adapters.common_adapter import CommonAdapter
from virtmat.middleware.exceptions import ResourceConfigurationError
from virtmat.middleware.resconfig import get_default_resconfig

res_mapper = {'nodes': 'nodes', 'ntasks_per_node': 'cpus_per_node',
              'cpus_per_task': 'cpus_per_task', 'walltime': 'time',
              'mem_per_cpu': 'memory_per_cpu'}


def get_default_qadapter(cfg=None, w_name=None, lp_file=None):
    """
    Construct a default qadapter from a resconfig object

    Args:
        cfg (ResConfig): rsource configuration object, if None then it will
        be loaded from resconfig file
        w_name (str): name of worker, if None then default_worker will be used
        lp_file (str): a custom launchpad file to be specified in the
        rocket_launch section of qadapter

    Returns:
        a CommonAdapter object if worker type is not None, otherwise None
    """
    if cfg is None:
        cfg = get_default_resconfig()
        if cfg is None:
            msg = 'Resource configuration file not found.'
            raise ResourceConfigurationError(msg)
    if w_name is None:
        wcfg = cfg.default_worker
        if wcfg is None:
            raise ResourceConfigurationError('no default worker found')
    else:
        try:
            wcfg = next(w for w in cfg.workers if w.name == w_name)
        except StopIteration as err:
            msg = f'no worker with {w_name} found'
            raise ResourceConfigurationError(msg) from err
    if wcfg.type_ is None:
        return None
    queue = wcfg.default_queue
    if queue is None:
        raise ResourceConfigurationError('no default queue found')
    qadapter = {}
    qadapter['q_type'] = wcfg.type_
    qadapter['q_name'] = wcfg.default_queue.name
    qadapter['queue'] = wcfg.default_queue.name  # the launcher requires this redundant key
    for key1, key2 in res_mapper.items():
        res = queue.get_resource(key2)
        if res and res.default:
            qadapter[key1] = res.default
    qadapter['ntasks'] = qadapter['nodes'] * qadapter['ntasks_per_node']
    if lp_file is not None:
        if not os.path.exists(lp_file):
            raise ResourceConfigurationError(f'file {lp_file} does not exist')
        qadapter['rocket_launch'] = f'rlaunch -l {lp_file} singleshot'
    else:
        qadapter['rocket_launch'] = 'rlaunch singleshot'
    qadapter['account'] = wcfg.default_account
    qadapter['pre_rocket'] = get_pre_rocket(wcfg)
    if wcfg.default_launchdir:
        qadapter['launch_dir'] = wcfg.default_launchdir
    return CommonAdapter(**qadapter)


def get_pre_rocket(wcfg, **kwargs):
    """Return a pre_rocket string for qadapter

    Args:
        wcfg (WorkerConfig): worker config
        kwargs (dict): required resources

    Raises:
        ResourceConfigurationError: in case of no match
    """
    pre_rocket_lst = []
    if wcfg.modules:
        pre_rocket_lst.append('module purge')
    pre_rocket_lst.extend([m.get_command(m.name) for m in wcfg.default_modules])
    if kwargs.get('modules'):
        for key, val in kwargs.get('modules').items():
            miter = (m.get_command(key, val) for m in wcfg.modules)
            try:
                match = next(m for m in miter if m is not None)
            except StopIteration as err:
                msg = f'no matching module {key}{val}'
                raise ResourceConfigurationError(msg) from err
            pre_rocket_lst.append(match)
    default_venv = []
    if wcfg.default_venv:
        if wcfg.default_venv['type'] == 'conda':
            venv_name = wcfg.default_venv['name']
            default_venv.append(f'conda activate {venv_name}')
        elif wcfg.default_venv['type'] == 'venv':
            venv_prefix = wcfg.default_venv['prefix']
            activate = os.path.join(venv_prefix, 'bin', 'activate')
            default_venv.append(f'source {activate}')
    pre_rocket_lst.extend(default_venv)
    default_vars = []
    for key in wcfg.default_envvars:
        if key not in wcfg.envvars:
            raise ResourceConfigurationError(f'default envvar {key} not in envvars')
        val = wcfg.envvars[key]
        comm = f'unset {key}' if val is None else f'export {key}={val}'
        default_vars.append(comm)
    pre_rocket_lst.extend(default_vars)
    if kwargs.get('envs'):
        for key, val in kwargs.get('envs').items():
            if key not in wcfg.default_envvars:
                if key in wcfg.envvars:
                    val = wcfg.envvars[key]
                comm = f'unset {key}' if val is None else f'export {key}={val}'
                pre_rocket_lst.append(comm)
    pre_rocket_lst.extend([c.get_command() for c in wcfg.default_commands])
    return '; '.join(pre_rocket_lst)


def validate_resources(wcfg, qcfg, g_name=None, **kwargs):
    """Validate required resources for a specific worker and queue

    Args:
        wcfg (WorkerConfig): worker configuration object
        qcfg (QueueConfig): queue configuration object
        g_name (str): group name to be used for accounting
        kwargs (dict): required resources

    Raises:
        ResourceConfigurationError: if the required resources are not valid
    """
    if qcfg.name not in (q.name for q in wcfg.queues):
        msg = f'queue {qcfg.name} not in the list of queues'
        raise ResourceConfigurationError(msg)
    if g_name is not None:
        if g_name not in wcfg.accounts:
            msg = f'group {g_name} not in the list of groups'
            raise ResourceConfigurationError(msg)
    for key, val in kwargs.items():
        if key in res_mapper:
            qcfg.validate_resource(res_mapper[key], val)
    get_pre_rocket(wcfg, **kwargs)


def get_worker_get_queue(cfg=None, w_name=None, q_name=None, g_name=None, **kwargs):
    """
    Return worker and queue matching a resource requirement

    Args:
        cfg (ResConfig): resource configuration object, if None then it will
        be loaded from resconfig file
        w_name (str): name of worker
        q_name (str): name of the queue, ignored if w_name is None
        g_name (str): group name to be used for accounting
        kwargs (dict): required resources

    Returns:
        tuple(WorkerConfig, QueueConfig): worker and queue in case of match,
        (None, None) otherwise

    Raises:
        ResourceConfigurationError: if resconfig (cfg) is not provided / found
    """
    cfg = cfg or get_default_resconfig()
    if cfg is None:
        msg = 'Resource configuration file not found.'
        raise ResourceConfigurationError(msg)
    if w_name is None:
        for wcfg in (cfg.default_worker, *cfg.workers):
            for queue in (wcfg.default_queue, *wcfg.queues):
                if queue:
                    try:
                        validate_resources(wcfg, queue, g_name, **kwargs)
                    except ResourceConfigurationError:
                        continue
                    return wcfg, queue
    else:
        try:
            wcfg = next(w for w in cfg.workers if w.name == w_name)
        except StopIteration:
            return None, None
        if q_name is None:
            for queue in (wcfg.default_queue, *wcfg.queues):
                if queue:
                    try:
                        validate_resources(wcfg, queue, g_name, **kwargs)
                    except ResourceConfigurationError:
                        continue
                    return wcfg, queue
        else:
            try:
                queue = next(q for q in wcfg.queues if q.name == q_name)
                validate_resources(wcfg, queue, g_name, **kwargs)
            except (StopIteration, ResourceConfigurationError):
                return None, None
            return wcfg, queue
    return None, None


def get_custom_qadapter(cfg=None, w_name=None, q_name=None, g_name=None,
                        lp_file=None, **kwargs):
    """
    Construct a custom queue adapter for the first matching resources
    always starting from the default resources

    Args:
        cfg (ResConfig): resource configuration object, if None then it will
        be loaded from resconfig file
        w_name (str): name of worker
        q_name (str): name of the queue, ignored if w_name is None
        g_name (str): group name to be used for accounting
        lp_file (str): a custom launchpad file to be specified in the
        rocket_launch section of qadapter

    Returns:
        tuple (str, CommonAdapter): worker name, qadapter object in case of match,
        (None, None) otherwise
    """
    wcfg, qcfg = get_worker_get_queue(cfg, w_name, q_name, g_name, **kwargs)
    if not (wcfg and qcfg):
        return None, None
    return wcfg.name, get_qadapter(wcfg, qcfg, g_name, lp_file, **kwargs)


def get_qadapter(wcfg, qcfg, g_name=None, lp_file=None, **kwargs):
    """
    Return a qadapter for specific worker, queue and optionally specified
    resources

    Args:
        wcfg (WorkerConfig): worker configuration object
        qcfg (QueueConfig): queue configuration object
        g_name (str): group name to be used for accounting
        lp_file (str): a custom launchpad file to be specified in the
        rocket_launch section of qadapter

    Returns:
        CommonAdapter: qadapter object
        None: if worker type not specified
    """
    if wcfg.type_ is None:
        return None
    qadapter = {}
    qadapter['q_type'] = wcfg.type_
    assert qcfg.name in (q.name for q in wcfg.queues)
    qadapter['q_name'] = qcfg.name
    qadapter['queue'] = qcfg.name  # the launcher requires this redundant key
    ntasks = kwargs.pop('ntasks', None)
    if ntasks is not None:
        if 'cpus_per_task' not in kwargs:
            kwargs['cpus_per_task'] = qcfg.get_resource('cpus_per_task').default
        if 'ntasks_per_node' not in kwargs:
            if 'nodes' not in kwargs:
                ncores = ntasks * kwargs['cpus_per_task']
                ncores_per_node = qcfg.get_resource('cpus_per_node').maximum
                kwargs['nodes'] = math.ceil(ncores/ncores_per_node)
            kwargs['ntasks_per_node'] = ntasks // kwargs['nodes']
    for key, val in kwargs.items():
        if key in res_mapper:
            qadapter[key] = val
    for key1, key2 in res_mapper.items():
        if key1 not in kwargs:
            res = qcfg.get_resource(key2)
            if res and res.default:
                qadapter[key1] = res.default
    if lp_file is not None:
        if not os.path.exists(lp_file):
            raise ResourceConfigurationError(f'file {lp_file} does not exist')
        qadapter['rocket_launch'] = f'rlaunch -l {lp_file} singleshot'
    else:
        qadapter['rocket_launch'] = 'rlaunch singleshot'
    qadapter['ntasks'] = qadapter['nodes'] * qadapter['ntasks_per_node']
    if g_name is not None:
        assert g_name in wcfg.accounts
        qadapter['account'] = g_name
    else:
        qadapter['account'] = wcfg.default_account
    qadapter['pre_rocket'] = get_pre_rocket(wcfg, **kwargs)
    if wcfg.default_launchdir:
        qadapter['launch_dir'] = wcfg.default_launchdir
    return CommonAdapter(**qadapter)



# pylint: disable=unused-argument
""" A graphical user interface for WFEngine based on ipywidgets"""
import os
import os.path
import time
import re
from pathlib import Path
from functools import partial
import json
from ruamel import yaml
import ipywidgets as widgets
from ipywidgets import HBox, VBox, Layout
from ipyfilechooser import FileChooser
from IPython.display import display, clear_output, JSON
from fireworks import LaunchPad
from fireworks.user_objects.queue_adapters.common_adapter import CommonAdapter
from fireworks.fw_config import LAUNCHPAD_LOC
from fireworks import Workflow
from prettytable import PrettyTable
from virtmat.middleware.engine.wfengine import WFEngine
from virtmat.middleware.query.wfquery import WFQuery
from virtmat.middleware.engine.wfengine_remote import WFEngineRemote
from virtmat.middleware.exceptions import ConfigurationException
from virtmat.middleware.exceptions import InvalidStateException
from virtmat.middleware.utilities import get_logger
from virtmat.middleware.resconfig import get_default_resconfig, get_resconfig_loc
from virtmat.middleware.resconfig import ResConfig, set_defaults_from_guess


# Getting the arguments with FileChooser and SelectMultiple
lp_file = FileChooser('')
qa_file = FileChooser('')
wf_file = FileChooser('')

style = {'description_width': 'initial'}

# Set default paths for file chooser
HOME = str(Path.home())
fireworks_cfg_dir = os.path.join(HOME, '.fireworks')
lp_file.show_hidden = True
if os.path.exists(fireworks_cfg_dir):
    lp_file.default_path = fireworks_cfg_dir
lp_file.default_filename = 'launchpad.yaml'
qa_file.show_hidden = True
if os.path.exists(fireworks_cfg_dir):
    qa_file.default_path = fireworks_cfg_dir
qa_file.default_filename = 'qadapter.yaml'
wf_file.default_path = os.path.join(HOME)
wf_file.default_filename = 'workflow.yaml'

# Show folder icons for file chooser
lp_file.use_dir_icons = True
qa_file.use_dir_icons = True

# Set multiple file filter patterns for config file
lp_file.filter_pattern = ['*.yaml', '*.json']
qa_file.filter_pattern = ['*.yaml', '*.json']
wf_file.filter_pattern = ['*.yaml', '*.json']

# Set titles for choosers
lp_file.title = 'Select a launchpad file'
qa_file.title = 'Select a qadapter file'
wf_file.title = 'Select a workflow file'

text_input_layout = Layout(width='100%')

configure_button = widgets.Button(tooltip='Create a new and manage running engine',
                                  description='Manage engine', style=style, icon='')
output_configure_button = widgets.Output()

lpad_button = widgets.Button(description='Load launchpad', style=style, icon='upload')
output_lpad_button = widgets.Output()

qadapter_button = widgets.Button(description='Load qadapter', style=style, icon='upload')
output_qadapter_button = widgets.Output()

resconfig_button = widgets.Button(tooltip='Create a new resconfig', style=style,
                                  description='Create resconfig', icon='',
                                  disabled=True)
output_resconfig_button = widgets.Output()

worker_select = widgets.Dropdown(options=[], value=None, description='Workers',
                                 disabled=True)

new_engine_button = widgets.Button(tooltip='New engine will be created',
                                   description='Create new engine', style=style, icon='cogs')
output_new_engine_button = widgets.Output()

dump_engine_button = widgets.Button(tooltip='Save the current engine into a file',
                                    description='Save engine', style=style, icon='download')

resume_engine_button = widgets.Button(tooltip='Load an engine from a file',
                                      description='Load engine', style=style, icon='upload')
output_saveload_engine_button = widgets.Output()

manage_launcher_button = widgets.Button(tooltip='Manage launcher threads',
                                        description='Manage launcher', style=style, icon='')
output_manage_launcher_button = widgets.Output()

start_button = widgets.Button(tooltip='Start a launcher thread in the background',
                              description='Start launcher', style=style, icon='play')
output_start_button = widgets.Output()

stop_button = widgets.Button(tooltip='Stop the launcher thread in the background',
                             description='Stop launcher', style=style, icon='power-off')
output_stop_button = widgets.Output()

manage_workflows_button = widgets.Button(tooltip='Manage workflows', description='Manage workflows',
                                         style=style, icon='')
output_manage_workflows_button = widgets.Output()

new_workflow_button = widgets.Button(tooltip='Add workflows to the engine',
                                     description='Add workflows', style=style, icon='plus')
output_new_workflow_button = widgets.Output()

add_workflow_button = widgets.Button(tooltip='Add workflows from query/file to the engine',
                                     description='Add', style=style, icon='upload')
output_add_workflow_button = widgets.Output()

remove_workflow_button = widgets.Button(tooltip='Remove workflows from the engine',
                                        description='Remove workflows', style=style, icon='trash')
output_remove_workflow_button = widgets.Output()

commit_remove_workflow_button = widgets.Button(tooltip='Remove selected workflows from the engine',
                                               description='Remove', style=style, icon='trash')
output_commit_remove_workflow_button = widgets.Output()

manage_nodes_button = widgets.Button(tooltip='Manage individual workflow nodes',
                                     description='Manage nodes', style=style, icon='')
output_manage_nodes_button = widgets.Output()

status_button = widgets.Button(tooltip='Show the status of all workflows', description='Status',
                               style=style, icon='eye')
nodes_status_button = widgets.Button(tooltip='Show the status of all nodes', description='Status',
                                     style=style, icon='eye')
output_status_button = widgets.Output()

status_button_detail = widgets.Button(tooltip='Show the status of selected nodes',
                                      description='Status', style=style, icon='search-plus')
status_button_detail_output = widgets.Output()

rerun_node_button = widgets.Button(tooltip='Rerun selected nodes', description='Rerun nodes',
                                   style=style, icon='redo')
output_rerun_node_button = widgets.Output()

cancel_launch_button = widgets.Button(tooltip='Cancel the launches of selected nodes',
                                      description='Cancel launches', style=style, icon='xmark')
output_cancel_launch_button = widgets.Output()

update_node_button = widgets.Button(tooltip='Updates the spec dictionaries of selected nodes',
                                    description='Update nodes', style=style, icon='pencil')
output_update_node_button = widgets.Output()

update_rerun_node_button = widgets.Button(tooltip='Updates the selected nodes and reruns them',
                                          description='Update & rerun nodes', style=style)
output_update_rerun_node_button = widgets.Output()

add_nodes_button = widgets.Button(tooltip='Add nodes to a workflow', description='Add nodes',
                                  style=style, icon='plus')
output_add_nodes_button = widgets.Output()

add_node_button = widgets.Button(tooltip='Add the node to the workflow', description='Add node',
                                 style=style, icon='upload')
output_add_node_button = widgets.Output()

query = widgets.Textarea(value='null', placeholder='{}', description='Query',
                         tooltip='PyMongo Query to select workflows (JSON format)',
                         layout=text_input_layout)

wf_query = widgets.VBox([
    widgets.Textarea(value=None, placeholder='{"nodes": {"$in": [1, 2]}, "name": "WF name"}',
                     tooltip='PyMongo Query to select workflows (JSON format)',
                     description='WF query', layout=text_input_layout),
    widgets.Textarea(value=None, placeholder='{"spec.some_value": 4}',
                     tooltip='PyMongo Query to select fireworks (JSON format)',
                     description='FW query', layout=text_input_layout)
])

engine_file = FileChooser('')
engine_file.default_path = '.'
engine_file.use_dir_icons = True
engine_file.filter_pattern = ['*.yaml', '*.json']
engine_file.title = 'WFEngine file'

spec_dict = widgets.Textarea(value=None, description='Update',
                             tooltip='PyMongo spec update query (JSON format)',
                             placeholder='{"key 1": "string", "key 2": [1, 2]}',
                             style={'description_width': 'initial'})

configure_engine_method = widgets.RadioButtons(
    options=['Save or load an existing engine',
             'Create a new engine from scratch'],
    value='Save or load an existing engine',
    description='Method',
    tooltip='Select a method to construct an engine'
)
configure_engine_method_output = widgets.Output()

add_workflow_method = widgets.RadioButtons(options=['From query', 'From file'],
                                           value='From query', description='How to add workflows?',
                                           tooltip='Select the method to add workflows')
output_add_workflow_method = widgets.Output()

pause_restart = widgets.RadioButtons(options=['pause', 'restart'], value='pause',
                                     description='Cancel policy',
                                     tooltip='Select the cancellation policy')

remote_cluster = widgets.Checkbox(value=False, description='Use a remote cluster',
                                  tooltip='Use a remote cluster to launch jobs', indent=False)
remote_cluster_output = widgets.Output()

user_name = widgets.Textarea(value='', placeholder='xy1234', description='Username',
                             tooltip='Username on the remote cluster', layout=text_input_layout)

host_name = widgets.Textarea(value='', placeholder='hostname.domainname', description='Hostname',
                             tooltip='Hostname of the remote cluster', layout=text_input_layout)

remote_conf = widgets.Textarea(value='', placeholder='module load xyz', description='Command',
                               tooltip='Remote configuration command', layout=text_input_layout)

launch_dir = widgets.Textarea(value='', placeholder='/absolute/path/to/launchdir',
                              description='Launchdir',
                              tooltip='Top-level directory in that nodes will be executed',
                              layout=text_input_layout)

unique_launchdir = widgets.Checkbox(value=True,
                                    description='Use unique launch directories',
                                    tooltip='Use unique launch directories')


def clear_consoleoutput():
    """Clear outputs"""
    console_outputs = [output_status_button, output_remove_workflow_button,
                       output_rerun_node_button, output_update_node_button,
                       output_update_rerun_node_button, output_add_nodes_button,
                       output_add_node_button, output_new_workflow_button,
                       output_add_workflow_method, output_add_workflow_button,
                       output_cancel_launch_button, output_stop_button, output_start_button,
                       output_saveload_engine_button]
    for b_out in console_outputs:
        with b_out:
            clear_output()


def clear_button_outputs():
    """Clear top buttons outputs"""
    button_outputs = [output_configure_button, output_manage_launcher_button,
                      output_manage_workflows_button, output_manage_nodes_button]
    for b_out in button_outputs:
        with b_out:
            clear_output()
    clear_consoleoutput()


def configure_button_clicked(bvar):
    """Configure button is clicked"""
    clear_button_outputs()
    with output_configure_button:
        clear_output()
        display(VBox([configure_engine_method,
                      configure_engine_method_output]))
        configure_engine_method_changed(bvar)


def manage_launcher_button_clicked(bvar):
    """manage launcher button is clicked"""
    clear_button_outputs()
    with output_manage_workflows_button:
        clear_output()
        display(HBox([start_button, stop_button]), output_start_button,
                output_stop_button)


def new_workflow_button_clicked(bvar):
    """new workflow button is clicked"""
    clear_consoleoutput()
    with output_new_workflow_button:
        clear_output()
        display(VBox([add_workflow_method, output_add_workflow_method]))
        add_workflow_method_changed(bvar)
        display(add_workflow_button, output_add_workflow_button)


def add_workflow_method_changed(bvar):
    """select the method to add workflows from radio buttons"""
    with output_add_workflow_button:
        clear_output()
    with output_add_workflow_method:
        clear_output()
        if add_workflow_method.value == 'From query':
            print('Using a query to add workflows')
            display(wf_query)
        if add_workflow_method.value == 'From file':
            print('Add workflows from a file')
            display(wf_file)


def remote_cluster_changed(bvar):
    """toggle the remote cluster checkbox"""
    with remote_cluster_output:
        clear_output()
        if remote_cluster.value:
            display(user_name, host_name, remote_conf)


def configure_engine_method_changed(bvar):
    """select engine configuration method from radio buttons"""
    with configure_engine_method_output:
        clear_output()
        if configure_engine_method.value == 'Save or load an existing engine':
            print('Save or load an existing engine')
            display(remote_cluster)
            display(engine_file)
            display(HBox([dump_engine_button, resume_engine_button]),
                    output_saveload_engine_button)
        if configure_engine_method.value == 'Create a new engine from scratch':
            print('Create a new engine from scratch')
            display(lp_file)
            display(lpad_button, output_lpad_button)
            display(resconfig_button, output_resconfig_button)
            display(worker_select)
            cfg = get_default_resconfig()
            if cfg is None:
                resconfig_button.disabled = False
                new_engine_button.disabled = True
            else:
                worker_select.options = [w.name for w in cfg.workers]
                worker_select.value = cfg.default_worker.name
                worker_select.disabled = False
            display(qa_file)
            display(qadapter_button, output_qadapter_button)
            display(query, launch_dir, unique_launchdir)
            display(VBox([remote_cluster, remote_cluster_output]))
            remote_cluster_changed(bvar)
            display(new_engine_button, output_new_engine_button)


def resconfig_button_clicked(bvar):
    """resconfig button is clicked"""
    with output_resconfig_button:
        resconfig_loc = get_resconfig_loc()
        if os.path.exists(resconfig_loc):
            return
        cfg = ResConfig.from_scratch()
        set_defaults_from_guess(cfg.default_worker)
        print('resconfig successfully created')
        cfg.to_file(resconfig_loc)
        print('resconfig successfully saved')
        resconfig_button.disabled = True
        worker_select.options = [w.name for w in cfg.workers]
        worker_select.value = cfg.default_worker.name
        worker_select.disabled = False
        new_engine_button.disabled = False


class WFEnginejupyter():
    """A class for construcing a GUI for FireWorks"""
    wfe = None
    jqadapter = None
    wf_id_select = None
    node_id_select = None
    rows_inputs = None
    rows_outputs = None
    job_category = None
    func_name = None
    size_inp = None
    size_out = None
    logger = get_logger(__name__)

    def __init__(self):
        """Load default launchpad and qadapter"""
        if LAUNCHPAD_LOC:
            try:
                self.jlaunchpad = LaunchPad.from_file(LAUNCHPAD_LOC)
            except FileNotFoundError as error:
                with output_lpad_button:
                    print('The default launchpad file is missing.')
                raise error
            except KeyError as error:
                with output_lpad_button:
                    print('The default lpad file is not correct.')
                raise error
            with output_lpad_button:
                print('Default lpad in ' + LAUNCHPAD_LOC + ' is loaded')

    def new_engine_button_clicked(self, bvar):
        """create new engine button is clicked"""
        with output_new_engine_button:
            clear_output()
        if self.wfe and self.wfe.thread and self.wfe.thread.is_alive():
            self.wfe.stop()
            with output_new_engine_button:
                print('Stopping the running launcher thread, please wait ...')
            self.wfe.thread.join()
            with output_new_engine_button:
                clear_output()
        else:
            query_dict = json.loads(query.value)
            launchdir = launch_dir.value or None
            if remote_cluster.value:
                if host_name.value == '' or user_name.value == '':
                    msg = 'host_name or user_name empty.'
                    raise ConfigurationException(msg)
                if self.jqadapter is None:
                    msg = 'qadapter must be defined in remote cluster mode'
                    raise ConfigurationException(msg)
                self.wfe = WFEngineRemote(launchpad=self.jlaunchpad,
                                          qadapter=self.jqadapter,
                                          wf_query=query_dict,
                                          host=host_name.value,
                                          user=user_name.value,
                                          conf=remote_conf.value,
                                          launchdir=launchdir,
                                          name=worker_select.value,
                                          unique_launchdir=unique_launchdir.value)
                with output_new_engine_button:
                    print('Engine is created for remote launcher')
            else:
                self.wfe = WFEngine(launchpad=self.jlaunchpad,
                                    qadapter=self.jqadapter,
                                    wf_query=query_dict,
                                    launchdir=launchdir,
                                    name=worker_select.value,
                                    unique_launchdir=unique_launchdir.value)
                with output_new_engine_button:
                    print('Engine is created for local launcher')

    def resume_engine_button_clicked(self, bvar):
        """resume engine button is clicked"""
        with output_saveload_engine_button:
            clear_output()
        if self.wfe and self.wfe.thread and self.wfe.thread.is_alive():
            self.wfe.stop()
            with output_saveload_engine_button:
                print('Stopping the running launcher thread, please wait ...')
            self.wfe.thread.join()
        eng_class = WFEngineRemote if remote_cluster.value else WFEngine
        if engine_file.value is None:
            with output_saveload_engine_button:
                print('Select an engine file first.')
            return
        try:
            self.wfe = eng_class.from_file(filename=engine_file.value)
        except (FileNotFoundError, IsADirectoryError) as error:
            with output_saveload_engine_button:
                print('Specify a path to an existing engine file.')
                print(f'Error: {error}')
            raise
        except (PermissionError, json.JSONDecodeError, yaml.YAMLError) as error:
            with output_saveload_engine_button:
                print(f'Error: {error}')
            raise
        except Exception:
            msg = 'An error with loading the engine file ocurred. Check the log for more details.'
            with output_saveload_engine_button:
                print(msg)
            raise
        with output_saveload_engine_button:
            print('An engine has been loaded from file', engine_file.value)

    def dump_engine_button_clicked(self, bvar):
        """dump the engine to file"""
        with output_saveload_engine_button:
            clear_output()
            if self.wfe:
                try:
                    if engine_file.value is None:
                        raise ValueError('Select an engine file first.')
                    self.wfe.to_file(filename=engine_file.value)
                except (FileNotFoundError, IsADirectoryError) as error:
                    print('Specify a valid engine filename or path.')
                    print(error.args)
                except ValueError as error:
                    if 'Unsupported format' in error.args[0]:
                        print('Specify a valid engine filename ending with .json or .yaml')
                        print(error.args)
                    else:
                        print(str(error))
                except PermissionError as error:
                    print(error.args)
                else:
                    print('The engine has been saved in file', engine_file.value)
            else:
                print('Create an engine first.')

    def lpad_button_clicked(self, bvar):
        """load user defined launchpad"""
        with output_lpad_button:
            clear_output()
            if lp_file.selected is None:
                print('Select a launchpad file.')
                time.sleep(1)
                return
            try:
                self.jlaunchpad = LaunchPad.from_file(lp_file.selected)
            except FileNotFoundError as error:
                print('Select a valid launchpad file')
                print(error.args)
            else:
                print('Launchpad has been loaded from file', lp_file.selected)

    def qadapter_button_clicked(self, bvar):
        """load user defined qadapter"""
        with output_qadapter_button:
            clear_output()
            if qa_file.selected is None:
                print('Select a qadapter file.')
                time.sleep(1)
                return
            try:
                self.jqadapter = CommonAdapter.from_file(qa_file.selected)
            except FileNotFoundError as error:
                print('Select a valid qadapter file')
                print(error.args)
            else:
                print('Qadapter has been loaded from file', qa_file.selected)

    def start_launcher_clicked(self, bvar):
        """start launcher button clicked"""
        clear_consoleoutput()
        try:
            self.wfe.start()
        except Exception as err:
            with output_start_button:
                print(f'Error occurred: {err}')
            raise
        with output_start_button:
            clear_output()
            print('The launcher thread is created.')

    def stop_launcher_clicked(self, bvar):
        """stop launcher button clicked"""
        clear_consoleoutput()
        with output_stop_button:
            clear_output()
            try:
                self.wfe.stop()
                print('please wait ...')
                self.wfe.thread.join()
                clear_output()
                print('Launcher thread is stopped.')
            except AttributeError:
                print('There is no running thread.')

    def manage_workflows_button_clicked(self, bvar):
        """manage workflows button is clicked"""
        clear_button_outputs()
        with output_manage_workflows_button:
            clear_output()
            display(HBox([status_button, new_workflow_button,
                          remove_workflow_button]), output_status_button,
                    output_new_workflow_button, output_remove_workflow_button)

    def status_button_clicked(self, bvar):
        """workflow status summary"""
        clear_consoleoutput()
        with output_status_button:
            clear_output()
            self.wfe.show_wf_status(add_io_info=False)

    def add_workflow_button_clicked(self, bvar):
        """add workflows from a query or a file"""
        with output_add_workflow_button:
            clear_output()
        if add_workflow_method.value == 'From file':
            if wf_file.selected is None:
                with output_add_workflow_button:
                    print('Select a workflow file first')
                return
            try:
                with open(wf_file.selected, 'rt', encoding='utf-8') as wff:
                    wf_obj = yaml.safe_load(wff)
            except FileNotFoundError as err:
                with output_add_workflow_button:
                    print('Specify a path to an existing workflow file.')
                    print(f'Error: {err}')
                raise
            except Exception:
                msg = 'An error occurred loading the workflow file. Check the logs for details.'
                with output_add_workflow_button:
                    print(msg)
                raise
            if isinstance(wf_obj, dict):
                wf_lst = [wf_obj]
            elif isinstance(wf_obj, list):
                wf_lst = wf_obj
            else:
                msg = 'The workflow file must contain either a dict or a list.'
                with output_add_workflow_button:
                    print(f'Error: {msg}')
                raise TypeError(msg)
            try:
                for wf_dct in wf_lst:
                    self.wfe.add_workflow(workflow=Workflow.from_dict(wf_dct))
            except Exception as err:
                with output_add_workflow_button:
                    print(f'Error: {err}')
                raise
            msg = f'{len(wf_lst)} workflow(s) have been loaded from file {wf_file.selected}'
            with output_add_workflow_button:
                print(msg)
        elif add_workflow_method.value == 'From query':
            try:
                wfq_dct = json.loads(wf_query.children[0].value)
                if not isinstance(wfq_dct, dict):
                    raise ValueError('The WF query must be a dictionary')
                fwq_dct = json.loads(wf_query.children[1].value)  # pylint: disable=E1136
                if not isinstance(fwq_dct, dict):
                    raise ValueError('The FW query must be a dictionary')
            except (ValueError, json.JSONDecodeError) as err:
                with output_add_workflow_button:
                    print(f'Error: {str(err)}')
                raise
            wfq = WFQuery(self.wfe.launchpad, wfq_dct, fwq_dct, metadata_only=True)
            with output_add_workflow_button:
                print(f'Adding the workflow IDs: {wfq.get_wf_ids()}')
            for wf_id in wfq.get_wf_ids():
                try:
                    self.wfe.add_workflow(fw_id=wf_id)
                except ConfigurationException as err:
                    with output_add_workflow_button:
                        print(f'{str(err)} {wf_id}')
                    raise
                with output_add_workflow_button:
                    print(f'Workflow {wf_id} has been loaded from query ')

    def create_wf_id_select(self):
        """creates a new selector with updated workflow ids"""
        self.wf_id_select = widgets.SelectMultiple(
            options=self.wfe.wf_ids,
            value=[],
            description='Workflow IDs',
            disabled=False
        )

    def remove_workflow_button_clicked(self, bvar):
        """remove workflows from engine"""
        clear_consoleoutput()
        with output_commit_remove_workflow_button:
            clear_output()
        self.create_wf_id_select()
        with output_remove_workflow_button:
            display(self.wf_id_select)
            display(commit_remove_workflow_button, output_commit_remove_workflow_button)

    def commit_remove_workflow_button_clicked(self, bvar):
        """commit workflows removal from engine"""
        with output_commit_remove_workflow_button:
            clear_output()
            for wf_id in self.wf_id_select.value:
                self.wfe.remove_workflow(wf_id)
            print(f'{len(self.wf_id_select.value)} workflow(s) removed')
        self.create_wf_id_select()
        with output_remove_workflow_button:
            clear_output()
            display(self.wf_id_select)
            display(commit_remove_workflow_button, output_commit_remove_workflow_button)

    def manage_nodes_button_clicked(self, bvar):
        """Manage nodes button is clicked"""
        self.node_id_select = widgets.SelectMultiple(
            options=self.wfe.fw_ids,
            value=[],
            description='Node IDs',
            disabled=False
        )
        clear_button_outputs()
        with output_manage_nodes_button:
            clear_output()
            display(HBox([self.node_id_select, spec_dict, pause_restart]))
            display(HBox([nodes_status_button, status_button_detail,
                          rerun_node_button, update_node_button,
                          update_rerun_node_button, cancel_launch_button,
                          add_nodes_button]),
                    output_status_button, status_button_detail_output,
                    output_rerun_node_button, output_update_node_button,
                    output_update_rerun_node_button, output_cancel_launch_button,
                    output_add_nodes_button)

    def nodes_status_button_clicked(self, bvar):
        """nodes status summary"""
        clear_consoleoutput()
        with output_status_button:
            clear_output()
            self.wfe.show_nodes_status()
        exp_time = 14400
        lost_runs = self.wfe.get_lost_jobs(time=exp_time)
        if len(lost_runs) != 0:
            with output_status_button:
                print(f'\nRunning nodes not been updated within {exp_time} s')
                print(lost_runs)

        unres_nodes = self.wfe.get_unreserved_nodes(time=exp_time)
        if len(unres_nodes) != 0:
            columns = unres_nodes[0].keys()
            table = PrettyTable(columns)
            for node in unres_nodes:
                table.add_row([node[c] for c in columns])
            with output_status_button:
                print(f'\nReserved nodes not been updated within {exp_time} s')
                print(table)

    def status_detailed_button_clicked(self, bvar):
        """status details about selected nodes"""
        clear_consoleoutput()
        lst = self.wfe.status_detail(*self.node_id_select.value)
        with output_status_button:
            display(JSON(lst))

    def rerun_node_button_clicked(self, bvar):
        """rerun selected nodes and print their new status"""
        clear_consoleoutput()
        for node in self.node_id_select.value:
            try:
                self.wfe.rerun_node(node)
            except InvalidStateException as err:
                with output_rerun_node_button:
                    print(err)
            except Exception as err:
                with output_rerun_node_button:
                    print('Unknown error. Check the logs.')
                raise err
            else:
                with output_rerun_node_button:
                    print(f'Node ID {node} has been rerun.')

    def update_node_button_clicked(self, bvar):
        """update selected nodes"""
        clear_consoleoutput()
        try:
            dct = json.loads(spec_dict.value)
        except json.JSONDecodeError as jerr:
            with output_update_node_button:
                print(jerr)
        else:
            for node in self.node_id_select.value:
                try:
                    self.wfe.update_node(node, dct)
                except InvalidStateException as err:
                    with output_update_node_button:
                        print(err)
                except Exception as err:
                    with output_update_node_button:
                        print('Unknown error. Check the logs.')
                        print(err)
                    raise err
                else:
                    with output_update_node_button:
                        print(f'Node ID {node} has been updated.')

    def update_rerun_node_button_clicked(self, bvar):
        """update and rerun selected nodes"""
        clear_consoleoutput()
        try:
            dct = json.loads(spec_dict.value)
        except json.JSONDecodeError as jerr:
            with output_update_rerun_node_button:
                print(jerr)
        else:
            for node in self.node_id_select.value:
                try:
                    self.wfe.update_rerun_node(node, dct)
                except InvalidStateException as err:
                    with output_update_rerun_node_button:
                        print(err)
                except Exception as err:
                    with output_update_rerun_node_button:
                        print('Unknown error. Check the logs.')
                        print(err)
                    raise err
                else:
                    with output_update_rerun_node_button:
                        print(f'Node ID {node} has been updated and rerun.')

    def add_nodes_button_clicked(self, bvar):
        """add nodes button is clicked"""
        clear_consoleoutput()

        def get_input_row():
            dropdown_kwargs = {'options': self.wfe.fw_ids, 'value': None,
                               'description': 'Node ID', 'disabled': False}
            while True:
                name_widget = widgets.Text(description='Name')
                value_widget = widgets.Text(description='Value', value='null')
                node_id_widget = widgets.Dropdown(**dropdown_kwargs)
                yield [name_widget, value_widget, node_id_widget]

        get_input_row_iter = get_input_row()

        def on_value_change(change, wids, wout, wtype=None):
            wout.clear_output()
            size = change['new']
            if size <= len(wids):
                wids = wids[:size]
            else:
                nadd = size - len(wids)
                if wtype == 'inputs':
                    elems = next(get_input_row_iter)
                    wids.extend([widgets.HBox(elems)]*nadd)
                elif wtype == 'outputs':
                    wids.extend([widgets.Text(description='Name')]*nadd)
            with wout:
                display(widgets.VBox(wids))

        self.rows_inputs = [widgets.HBox(next(get_input_row_iter))]
        output_inputs = widgets.Output()
        self.rows_outputs = [widgets.Text(description='Name')]
        output_outputs = widgets.Output()
        radio_kwargs = {'options': ['interactive', 'batch', 'remote'],
                        'value': 'interactive', 'description': 'Job category',
                        'tooltip': 'Select job category'}
        self.job_category = widgets.RadioButtons(**radio_kwargs)
        self.func_name = widgets.Text(description='Function', value=None, placeholder='math.log',
                                      tooltip='Fully qualified name of a Python function')
        self.size_inp = widgets.BoundedIntText(value=1, min=0, description='# Inputs')
        self.size_out = widgets.BoundedIntText(value=1, min=0, description='# Outputs')
        self.size_inp.observe(partial(on_value_change, wids=self.rows_inputs,
                              wout=output_inputs, wtype='inputs'), 'value')
        self.size_out.observe(partial(on_value_change, wids=self.rows_outputs,
                              wout=output_outputs, wtype='outputs'), 'value')
        with output_add_nodes_button:
            display(self.func_name)
            display(self.size_inp)
            display(output_inputs)
            display(self.size_out)
            display(output_outputs)
            display(self.job_category)
            display(add_node_button, output_add_node_button)
            with output_inputs:
                display(widgets.VBox(self.rows_inputs))
            with output_outputs:
                display(widgets.VBox(self.rows_outputs))

    def add_node_button_clicked(self, bvar):
        """add node button is clicked"""
        output_add_node_button.clear_output()
        id_pattern = r'[^\d\W]\w*\b'
        inputs = []
        outputs = []
        try:
            if not re.match(r'^'+id_pattern+r'(\.'+id_pattern+r')*$', self.func_name.value):
                raise ValueError(f'Invalid function name: {self.func_name.value}')
            for row in self.rows_inputs[:self.size_inp.value]:
                input_name = row.children[0].value
                if not re.match(r'^'+id_pattern+r'$', input_name):
                    raise ValueError(f'Input name not valid: {input_name}')
                input_val = json.loads(row.children[1].value)
                input_node_id = row.children[2].value
                inputs.append((input_node_id, input_name, input_val))
            for row in self.rows_outputs[:self.size_out.value]:
                output_val = row.value
                if not re.match(r'^'+id_pattern+r'$', output_val):
                    raise ValueError(f'Output name not valid: {output_val}')
                outputs.append(output_val)
        except (ValueError, json.JSONDecodeError) as err:
            with output_add_node_button:
                print(f'Error: {str(err)}')
            raise err
        self.wfe.add_node(func=self.func_name.value, inputs=inputs,
                          outputs=outputs, category=self.job_category.value)
        with output_add_node_button:
            print('Added a new node.')

    def cancel_launch_button_clicked(self, bvar):
        """cancel launched (reserved or running) nodes"""
        clear_consoleoutput()
        with output_cancel_launch_button:
            clear_output()
        cancel_kwargs = {pause_restart.value: True}
        for fw_id in self.node_id_select.value:
            try:
                self.wfe.cancel_job(fw_id, **cancel_kwargs)
            except (InvalidStateException, ConfigurationException) as err:
                with output_cancel_launch_button:
                    print(err)
            except Exception as err:
                with output_cancel_launch_button:
                    print('Unknown error. Check the logs.', err)
                raise
            else:
                with output_cancel_launch_button:
                    print(f'Launch of node {fw_id} has been cancelled.')


wfengine = WFEnginejupyter()

new_engine_button.on_click(wfengine.new_engine_button_clicked)
resume_engine_button.on_click(wfengine.resume_engine_button_clicked)
resconfig_button.on_click(resconfig_button_clicked)
configure_button.on_click(configure_button_clicked)
new_workflow_button.on_click(new_workflow_button_clicked)
lpad_button.on_click(wfengine.lpad_button_clicked)
qadapter_button.on_click(wfengine.qadapter_button_clicked)
start_button.on_click(wfengine.start_launcher_clicked)
stop_button.on_click(wfengine.stop_launcher_clicked)
status_button.on_click(wfengine.status_button_clicked)
nodes_status_button.on_click(wfengine.nodes_status_button_clicked)
status_button_detail.on_click(wfengine.status_detailed_button_clicked)
dump_engine_button.on_click(wfengine.dump_engine_button_clicked)
manage_launcher_button.on_click(manage_launcher_button_clicked)
manage_workflows_button.on_click(wfengine.manage_workflows_button_clicked)
add_workflow_button.on_click(wfengine.add_workflow_button_clicked)
remove_workflow_button.on_click(wfengine.remove_workflow_button_clicked)
commit_remove_workflow_button.on_click(wfengine.commit_remove_workflow_button_clicked)
manage_nodes_button.on_click(wfengine.manage_nodes_button_clicked)
rerun_node_button.on_click(wfengine.rerun_node_button_clicked)
update_node_button.on_click(wfengine.update_node_button_clicked)
update_rerun_node_button.on_click(wfengine.update_rerun_node_button_clicked)
add_nodes_button.on_click(wfengine.add_nodes_button_clicked)
add_node_button.on_click(wfengine.add_node_button_clicked)
cancel_launch_button.on_click(wfengine.cancel_launch_button_clicked)
remote_cluster.observe(remote_cluster_changed)
configure_engine_method.observe(configure_engine_method_changed)
add_workflow_method.observe(add_workflow_method_changed)

display(HBox([configure_button, manage_launcher_button,
              manage_workflows_button, manage_nodes_button]),
        output_configure_button, output_manage_launcher_button,
        output_manage_workflows_button, output_manage_nodes_button)



"""command-line tools to configure available resources"""
import os
from virtmat.middleware.resconfig import get_resconfig_loc, configure


def setup_resconfig(resconfig_loc=None):
    """setup resource configuration interactively"""
    if resconfig_loc is None:
        resconfig_loc = get_resconfig_loc()
    if os.path.exists(resconfig_loc):
        print(f'Found resource configuration file {resconfig_loc}.')
        return

    print(f'Resource configuration file {resconfig_loc} not found.\n'
          'Do you want to create it?')
    while True:
        try:
            inp = input('Yes(default) | No | Ctrl+C to skip: ').strip()
        except KeyboardInterrupt:
            print('\n')
            break
        else:
            if inp.lower() in ('no', 'n'):
                break
            if inp.lower() in ('yes', ''):
                configure(resconfig_loc)
                break
            continue



"""various utilities for the VRE middleware"""
import os
import logging
import time
import json
import subprocess as sp
from fireworks import Launch
from virtmat.middleware.exceptions import SlurmError, ConfigurationException

LOGGING_LEVELS = ('NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL')


def get_logger(name, default_level=logging.ERROR):
    """return a custom logger object"""
    logger = logging.getLogger(name)
    if not logger.hasHandlers():
        fstr = '%(asctime)s %(levelname)s %(name)s: %(message)s'
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(fstr))
        logger.addHandler(handler)
        logger.setLevel(default_level)
    return logger


def get_slurm_job_state(*args, method='json', **kwargs):
    """select the method to get the SLURM job state"""
    if method == 'parse':
        return get_slurm_job_state_parse(*args, **kwargs)
    if method == 'json':
        return get_slurm_job_state_json(*args, **kwargs)
    raise ConfigurationException(f'unknown method {method}')


def get_command_output(command):
    """run a slurm command (a string or list) and return the output"""
    try:
        output = sp.check_output(command)  # nosec B603
    except (sp.SubprocessError, OSError) as exc:
        if not isinstance(command, str):
            assert isinstance(command, list)
            command = ' '.join(command)
        raise SlurmError(f'error executing \"{command}\": {str(exc)}') from exc
    return output.decode()


def get_slurm_job_state_json(res_id, max_trials=3, trial_delay=3, trial=0):
    """extract the SLURM job state using the sacct command"""
    if trial > max_trials:
        msg = f'cannot get SLURM job state after {max_trials} trials'
        raise SlurmError(msg)
    command = ['sacct', '-j', str(res_id), '--json']
    try:
        jobs = json.loads(get_command_output(command))['jobs']
    except (json.JSONDecodeError, KeyError, TypeError) as exc:
        raise SlurmError('problem with collecting sacct command output') from exc
    if len(jobs) == 0:
        time.sleep(trial_delay)
        return get_slurm_job_state(res_id, max_trials, trial_delay, trial+1)
    assert jobs[0]['job_id'] == int(res_id)
    assert isinstance(jobs[0]['state']['current'], list)
    assert len(jobs[0]['state']['current']) == 1
    return jobs[0]['state']['current'][0]


def get_slurm_job_state_parse(res_id, max_trials=10, trial_delay=3, trial=0):
    """extract the SLURM job state using the sacct command"""
    if trial > max_trials:
        raise SlurmError(f'cannot get SLURM job state after {max_trials} trials')
    command = ['sacct', '-j', str(res_id), '-n', '-o', 'state']
    state = get_command_output(command).strip()
    if len(state) == 0:
        time.sleep(trial_delay)
        return get_slurm_job_state(res_id, max_trials, trial_delay, trial+1)
    return state


def await_slurm_job_state(res_id, state, sleep_time=10):
    """return when a specified SLURM job state has been reached"""
    while get_slurm_job_state(res_id) != state:
        time.sleep(sleep_time)


def lmod_env_module(command):
    """Process lmod modules list and spider commands and return modules

    Due to ambiguous format of provided input it is assumed that the last
    division is the module version in the case of more than one division.

    Args:
        command (str): either 'list' or 'spider'

    Returns:
        (dict): {prefix/.../name: [version1, ...]} in case of more divisions
                {'name': []} in case of one division

    """
    assert isinstance(command, str) and command in ('list', 'spider')
    lmod_cmd = os.environ.get('LMOD_CMD')
    if lmod_cmd is None:
        return None
    full_command = [lmod_cmd, 'bash', '-t', command]
    with sp.Popen(full_command, stdout=sp.DEVNULL, stderr=sp.PIPE) as proc:  # nosec B603
        string = proc.stderr.read().decode()
    if string.strip() == 'No modules loaded':
        return None
    mods_lst = string.strip().split('\n')
    mods_dct = {}
    for mod in mods_lst:
        divisions = mod.split('/')
        name = '/'.join(divisions[:-1]) if len(divisions) > 1 else divisions[0]
        mods_dct[name] = []
    for mod in mods_lst:
        divisions = mod.split('/')
        if len(divisions) > 1:
            version = divisions[-1]
            if version:
                mods_dct['/'.join(divisions[:-1])].append(version)
    return mods_dct


def format_warning_wrapper(func):
    """format user warnings, leave python warnings unchanged"""
    def wrapper(*args, **kwargs):
        warning = args[0]
        warning_cls = args[1]
        if isinstance(warning, UserWarning):
            assert warning_cls is UserWarning
            return 'Warning: ' + str(warning) + '\n'
        return func(*args, **kwargs)
    return wrapper


def exec_cancel(res_id):
    """
    Execute the Slurm cancel command

    Args:
        res_id (int): reservation ID (=Slurm JobID)

    Raises:
        SlurmError: if the command returns an error code
    """
    logger = get_logger(__name__)
    get_command_output(['/usr/bin/scancel', res_id])
    new_state = get_slurm_job_state(res_id)
    if new_state != 'CANCELLED':  # not covered
        msg = f'job could not be cancelled, new state {new_state}'
        logger.error(msg)
        raise RuntimeError(msg)
    logger.info('Cancelled res_id: %d', res_id)


def get_unreserved_nodes(launchpad, fw_ids, expiration_time=1209600):
    """
    Detect reserved nodes, i.e. in 'RESERVED' state within FireWorks, that
    have not been updated for a while. Possible inconsistent states in
    SLURM are 'CANCELLED', 'FAILED', 'COMPLETED', 'OUT_OF_MEMORY',
    'BOOT_FAIL', 'TIMEOUT' and 'DEADLINE'

    Args:
        launchpad (LaunchPad): launchpad object
        fw_ids ([int]): firework IDs
        expiration_time (int): minimum time in seconds since the most recent update,
            default is 1209600 seconds

    Returns:
        ([dict]): a list of dictionaries containing the firework ids, the
        reservation ids, the SLURM states and the launch directories of
        such reserved nodes
    """
    launch_ids = launchpad.detect_unreserved(expiration_secs=expiration_time)
    l_q = {'launch_id': {'$in': launch_ids}, 'fw_id': {'$in': fw_ids}}
    l_p = {'action': False, 'trackers': False}
    launches = [Launch.from_dict(d) for d in launchpad.launches.find(l_q, l_p)]
    res_ids = [launchpad.get_reservation_id_from_fw_id(i.fw_id) for i in launches]
    states = []
    for res_id in res_ids:
        try:
            state = get_slurm_job_state(res_id)
        except SlurmError:  # not covered
            state = None
        states.append(state)
    launch_dirs = [i.launch_dir for i in launches]
    items = zip([i.fw_id for i in launches], res_ids, states, launch_dirs)
    keys = ('fw_id', 'res_id', 'slurm_state', 'launch_dir')
    return [dict(zip(keys, values)) for values in items]



"""run a simple workflow engine from the command line"""
import json
import signal
from time import sleep
from argparse import ArgumentParser, BooleanOptionalAction
from fireworks import LaunchPad, Workflow
from fireworks.fw_config import LAUNCHPAD_LOC
from fireworks.user_objects.queue_adapters.common_adapter import CommonAdapter
from virtmat.middleware.engine.wfengine import WFEngine
from virtmat.middleware.utilities import logging, get_logger, LOGGING_LEVELS
from virtmat.middleware.exceptions import ConfigurationException

EXIT_NOW = False


def exit_gracefully(signum, _):
    """signal handler callback function"""
    globals()['EXIT_NOW'] = True
    logger = get_logger('run_wfengine', getattr(logging, 'INFO'))
    logger.info('signal %s received, please wait to exit ...', signum)


signal.signal(signal.SIGINT, exit_gracefully)
signal.signal(signal.SIGTERM, exit_gracefully)


def parse_clargs():
    """parse the command line arguments for the script"""
    m_description = 'run a simple workflow engine from the command line'
    parser = ArgumentParser(description=m_description)
    add_arguments(parser)
    return parser.parse_args()


def add_arguments(parser):
    """add arguments to a parser object"""
    parser.add_argument('-l', '--launchpad-file', default=LAUNCHPAD_LOC,
                        help='path to launchpad file')
    parser.add_argument('-q', '--qadapter-file', default=None, required=False,
                        help='path to default qadapter file')
    parser.add_argument('-n', '--worker-name', default=None, required=False,
                        help='worker name')
    parser.add_argument('-d', '--launchdir', default=None, required=False,
                        help='path to launch directory')
    parser.add_argument('-s', '--sleep-time', default=30, required=False,
                        type=int, help='sleep time for background evaluation in seconds')
    parser.add_argument('-c', '--category', default='all', required=False,
                        choices=('all', 'interactive', 'batch'),
                        help='category of nodes to run')
    parser.add_argument('-f', '--workflow-file', required=False, default=None,
                        help='path to a workflow file')
    parser.add_argument('--wf-query', type=lambda x: x if x else None, default=None,
                        required=False, help='workflow query as JSON string')
    parser.add_argument('--unique-launchdir', default=True, required=False,
                        action=BooleanOptionalAction, help='use unique launchdir')
    parser.add_argument('-r', '--autorun', default=False, required=False,
                        action='store_true', help='run workflow nodes')
    parser.add_argument('--loop', default=True, action=BooleanOptionalAction,
                        required=False, help='run nodes in an endless loop')
    parser.add_argument('--load-from-file', default=None, required=False,
                        type=str, help='load a wfengine from file')
    parser.add_argument('--dump-to-file', default=None, type=str,
                        required=False, help='dump the wfengine to file')
    parser.add_argument('--enable-logging', default=False, required=False,
                        action='store_true', help='enable logging messages')
    parser.add_argument('--logging-level', default='CRITICAL', required=False,
                        choices=LOGGING_LEVELS, help='logging level')


def wfengine_main(clargs):
    """the main function taking clargs as parameter"""
    if not clargs.enable_logging:
        logging.disable(logging.CRITICAL)

    logger = get_logger('run_wfengine', getattr(logging, clargs.logging_level))
    wf_query = clargs.wf_query and json.loads(clargs.wf_query)

    if clargs.load_from_file:
        logger.info('wfengine file: %s', clargs.load_from_file)
        wfe = WFEngine.from_file(clargs.load_from_file)
        if wf_query is not None:
            wfe.wf_query = wf_query
    elif clargs.launchpad_file:
        logger.info('launchpad file: %s', clargs.launchpad_file)
        launchpad = LaunchPad.from_file(clargs.launchpad_file)
        logger.info('qadapter file: %s', clargs.qadapter_file)
        qadapter = clargs.qadapter_file and CommonAdapter.from_file(clargs.qadapter_file)
        wfe = WFEngine(launchpad, qadapter=qadapter, wf_query=wf_query,
                       name=clargs.worker_name, launchdir=clargs.launchdir,
                       unique_launchdir=clargs.unique_launchdir,
                       sleep_time=clargs.sleep_time)
    else:
        logger.critical('no launchpad could be detected')
        raise ConfigurationException('no launchpad could be detected')

    logger.info('workflow query: %s', wfe.wf_query)
    logger.info('worker name: %s', wfe.name)
    logger.info('launch directory: %s', wfe.launchdir)
    logger.info('unique launch directory: %s', wfe.unique_launchdir)
    logger.info('workflow file: %s', clargs.workflow_file)

    if clargs.workflow_file:
        wfe.add_workflow(Workflow.from_file(clargs.workflow_file))

    if clargs.dump_to_file:
        logger.info('dumping engine to file: %s', clargs.dump_to_file)
        wfe.to_file(clargs.dump_to_file)

    logger.info('autorun: %s', clargs.autorun)

    launcher = {'batch': 'qlaunch', 'interactive': 'rlaunch'}
    categories = ('batch', 'interactive') if clargs.category == 'all' else (clargs.category,)

    if clargs.autorun:
        logger.info('loop: %s', clargs.loop)
        logger.info('sleep time: %s', wfe.sleep_time)
        logger.info('running categories: %s', categories)
        logger.info('to exit, press Ctrl+C and wait %s seconds', wfe.sleep_time)
        while True:
            for category in categories:
                fw_q = {'state': 'READY', 'spec._category': category}
                fw_ids = wfe.launchpad.get_fw_ids_in_wfs(wfe.wf_query, fw_q)
                logger.debug('%s nodes to run: %s', category, fw_ids)
                for fw_id in fw_ids:
                    getattr(wfe, launcher[category])(fw_id)
            logger.info('failed nodes: %s', wfe.get_failed())
            sleep(wfe.sleep_time)
            if EXIT_NOW or not clargs.loop:
                break


def main():
    """main function in module used as entry point"""
    wfengine_main(parse_clargs())



"""A simple workflow engine"""
import threading
import os
from time import sleep
from fireworks import fw_config, LaunchPad, Workflow, Firework, PyTask, FWorker
from fireworks.core.rocket_launcher import launch_rocket
from fireworks.queue.queue_launcher import launch_rocket_to_queue
from fireworks.utilities.fw_utilities import explicit_serialize
from fireworks.utilities.fw_serializers import FWSerializable
from fireworks.utilities.fw_serializers import serialize_fw
from fireworks.utilities.fw_serializers import recursive_serialize
from fireworks.utilities.fw_serializers import recursive_deserialize
from fireworks.utilities.fw_utilities import get_fw_logger
from fireworks.utilities.fw_utilities import create_datestamp_dir
import fireworks_schema
from prettytable import PrettyTable
from virtmat.middleware.query.wfquery import WFQuery
from virtmat.middleware.resconfig import get_default_resconfig
from virtmat.middleware.resconfig import get_default_qadapter
from virtmat.middleware.exceptions import ConfigurationException
from virtmat.middleware.exceptions import InvalidStateException
from virtmat.middleware.exceptions import ResourceConfigurationError
from virtmat.middleware.utilities import get_logger, get_unreserved_nodes
from virtmat.middleware.utilities import exec_cancel

SCHEMA = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'wfengine.json')
fireworks_schema.register_schema(SCHEMA)


@explicit_serialize
class WFEngine(FWSerializable):
    """A simple engine to manage workflows"""
    logger = get_logger(__name__)

    def __init__(self, launchpad, qadapter=None, wf_query=None, name=None,
                 launchdir=None, unique_launchdir=True, sleep_time=30):
        """
        Args:
            launchpad (LaunchPad): a launchpad to establish connection to a
                FireWorks database
            qadapter (CommonAdapter, None): a qadapter for submitting
                batch jobs
            wf_query (dict, None): a workflow query
            name (str, None): name of the engine; it must be a valid worker name;
                if None then the default worker name will be used
            launchdir (str): top-level launch directory for both interactive
                and batch jobs; if None then default_launchdir from the resconfg
                is used if not None, otherwise the current working directory
            unique_launchdir (bool): create individual directory with unique
                name for every single launch of interactive nodes, default True
            sleep_time (int): launcher thread awakes every `sleep_time` seconds
        """
        if isinstance(launchpad, LaunchPad):
            self.launchpad = launchpad
        else:
            msg = f'launchpad is not of type LaunchPad but {type(launchpad)}'
            self.logger.error(msg)
            raise TypeError(msg)

        self.wf_ids = ([] if wf_query is None else
                       self.launchpad.get_wf_ids(wf_query))
        self.unique_launchdir = unique_launchdir
        self.sleep_time = sleep_time
        self.wf_query = {'nodes': {'$in': self.wf_ids}}
        self._set_name(name)
        self._set_launchdir(launchdir)
        self._set_qadapter(qadapter)
        self._set_fworkers()
        self.nodes_torun = None
        self.thread = None
        self.event = None

    def _set_name(self, name=None):
        """set the name of the worker"""
        cfg = get_default_resconfig()
        if cfg is None:
            raise ConfigurationException('no resconfig found')
        if name is None:
            if cfg.default_worker is None:
                raise ConfigurationException('default worker not found')
            self.__name = cfg.default_worker.name
        else:
            if name not in (w.name for w in cfg.workers):
                msg = f'worker {name} not found in list of workers'
                raise ConfigurationException(msg)
            self.__name = name

    def _set_launchdir(self, launchdir=None):
        """set the launch directory"""
        if launchdir is None:
            cfg = get_default_resconfig()
            if cfg is None:
                raise ConfigurationException('no resconfig found')
            if cfg.default_worker is None:
                raise ConfigurationException('default worker not found')
            self.launchdir = cfg.default_worker.default_launchdir or os.getcwd()
        else:
            self.launchdir = launchdir
        if not os.path.exists(self.launchdir):
            msg = f'Launch directory {self.launchdir} does not exist'
            raise ConfigurationException(msg)

    def _set_qadapter(self, qadapter=None):
        """set the default qadapter"""
        assert self.name is not None
        if qadapter is None:
            try:
                self.qadapter = get_default_qadapter(w_name=self.name)
            except ResourceConfigurationError:
                self.qadapter = None
        else:
            self.qadapter = qadapter

    def _set_fworkers(self):
        """set the fworkers"""
        assert self.name is not None
        self.fworker_rlaunch = FWorker(name=self.name, category='interactive')
        if self.qadapter:
            self.fworker_qlaunch = FWorker(name=self.name, category='batch')
        else:
            self.fworker_qlaunch = None

    @property
    def name(self):
        """get the name of the engine"""
        return self.__name

    @name.setter
    def name(self, new_name):
        """set the name of the engine"""
        self._set_name(new_name)
        self._set_qadapter()
        self._set_fworkers()

    @property
    def wf_query(self):
        """get the query for the engine"""
        return self.__wf_query

    @wf_query.setter
    def wf_query(self, new_wf_query):
        """set the query for the engine"""
        self.__wf_query = new_wf_query
        self.__wf_ids = self.launchpad.get_wf_ids(self.wf_query)

    @property
    def wf_ids(self):
        """get the workflow ids of the engine"""
        return self.__wf_ids

    @wf_ids.setter
    def wf_ids(self, wf_ids):
        """set the workflow ids of the engine; wf_ids: a list of new wf_ids"""
        self.__wf_ids = wf_ids
        self.__wf_query = {'nodes': {'$in': self.wf_ids}}

    def append_wf_id(self, wf_id):
        """append a workflow id (wf_id) to the list of wf_ids"""
        self.__wf_ids.append(wf_id)
        self.__wf_query = {'nodes': {'$in': self.wf_ids}}

    def remove_wf_id(self, wf_id):
        """remove a workflow id (wf_id) from the list of wf_ids"""
        self.__wf_ids.remove(wf_id)
        self.__wf_query = {'nodes': {'$in': self.wf_ids}}

    @property
    def fw_ids(self):
        """get the current firework ids of the engine"""
        return self.launchpad.get_fw_ids_in_wfs(self.wf_query)

    def show_nodes_status(self):
        """Display the status summary of the nodes"""
        wfq = WFQuery(self.launchpad, wf_query=self.wf_query)
        fw_info = wfq.get_fw_info()
        if len(fw_info) > 0:
            columns = fw_info[0].keys()
            table = PrettyTable(columns)
            for fwk in fw_info:
                table.add_row([fwk[i] for i in columns])
            print(table)
        else:
            print('No nodes')

    def show_wf_status(self, add_io_info=True):
        """Display the status summary of the workflows"""
        wfq = WFQuery(self.launchpad, wf_query=self.wf_query)
        wfq.get_wf_info(add_io_info=add_io_info)

    def status_summary(self):
        """Display a status summary of workflows and nodes"""
        self.show_wf_status(add_io_info=False)
        print('Nodes summary:')
        self.show_nodes_status()

    def show_launcher_status(self):
        """Check whether a launcher thread is running"""
        if self.thread:
            if self.thread.is_alive():
                print('launcher thread is currently running')
            else:
                print('launcher thread not started')
        else:
            print('launcher thread not created')

    def status_detail(self, *fw_ids):
        """
        Print a detailed status of specified nodes

        Args:
            fw_ids ([int]): One or more fw_ids of the nodes
        Returns:
            a list of dictionaries, containing the nodes
        """
        def get_launches(launch_ids):
            launches = []
            projection = {'_id': False}
            for launch_id in launch_ids:
                l_query = {'launch_id': launch_id}
                launch = self.launchpad.launches.find_one(l_query, projection)
                launches.append(launch)
            return launches

        fw_list = []
        for fw_id in fw_ids:
            fw_query = {'fw_id': fw_id}
            projection = {'_id': False}
            fw_dict = self.launchpad.fireworks.find_one(fw_query, projection)
            fw_dict['launches'] = get_launches(fw_dict['launches'])
            fw_dict['archived_launches'] = get_launches(fw_dict['archived_launches'])
            fw_list.append(fw_dict)
        return fw_list

    def get_failed(self):
        """
        Get failed job ids

        Returns:
            ([int]): a list of fw_ids of failed jobs
        """
        wfq = WFQuery(self.launchpad, wf_query=self.wf_query,
                      fw_query={'state': 'FIZZLED'})
        return wfq.get_fw_ids()

    def qlaunch(self, fw_id):
        """
        Launch a batch node by submitting a job to the queuing system

        Args:
            fw_id (int): a fwd_id of the node to launch
        """

        if fw_config.MONGOMOCK_SERVERSTORE_FILE is not None:
            raise ConfigurationException('qlaunch error: cannot launch with Mongomock')
        if fw_id not in self.fw_ids:
            raise ConfigurationException(f'qlaunch error: invalid fw_id: {fw_id}')
        if self.qadapter is None:
            raise ConfigurationException('qlaunch error: qadapter is not defined')
        if self.fworker_qlaunch is None:
            raise ConfigurationException('qlaunch error: fworker is not defined')
        launch_rocket_to_queue(self.launchpad, self.fworker_qlaunch,
                               self.qadapter, reserve=True,
                               create_launcher_dir=True,
                               launcher_dir=self.launchdir,
                               fw_id=fw_id)

    def rlaunch(self, fw_id):
        """
        Launch an interactive node

        Args:
            fw_id (int): a fwd_id of the node to launch
        """
        if fw_id not in self.fw_ids:
            raise ConfigurationException(f'rlaunch error: invalid fw_id: {fw_id}')
        init_dir = os.getcwd()
        if self.unique_launchdir:
            logger = get_fw_logger('interactive', l_dir=self.launchpad.get_logdir())
            launch_dir = create_datestamp_dir(self.launchdir, logger, prefix='launcher_')
        else:
            launch_dir = self.launchdir
        try:
            os.chdir(launch_dir)
            launch_rocket(self.launchpad, self.fworker_rlaunch, fw_id)
        finally:
            os.chdir(init_dir)

    def get_lost_jobs(self, time=14400, fizzle=False):
        """
        Detect nodes that have been launched but not updated within the
        specified time.

        Args:
            time (int): minimim time in seconds since the most recent update
            fizzle (bool): set state of detected nodes to FIZZLED

        Returns:
            lost_fw_ids ([int]): a list of fw_ids of the lost runs
        """
        fw_query = {'fw_id': {'$in': self.fw_ids}}
        lostjobs = self.launchpad.detect_lostruns(expiration_secs=time,
                                                  fizzle=fizzle, query=fw_query)
        lost_fw_ids = lostjobs[1]
        if len(lost_fw_ids) != 0:
            self.logger.warning('Lost jobs detected: %s.', str(lost_fw_ids))
        return lost_fw_ids

    def get_unreserved_nodes(self, time=1209600):
        """
        Detect reserved nodes, i.e. in 'RESERVED' state within FireWorks, that
        have not been updated for a while. Possible inconsistent states in
        SLURM are 'CANCELLED', 'FAILED', 'COMPLETED', 'OUT_OF_MEMORY',
        'BOOT_FAIL', 'TIMEOUT' and 'DEADLINE'

        Args:
            time (int): minimum time in seconds since the most recent update

        Returns:
            ([dict]): a list of dictionaries containing the fw_ids, the
            reservation ids, the SLURM states and the launch directories of
            such reserved nodes
        """
        return get_unreserved_nodes(self.launchpad, self.fw_ids, expiration_time=time)

    @serialize_fw
    @recursive_serialize
    def to_dict(self):
        return {'launchpad': self.launchpad,
                'qadapter': self.qadapter,
                'wf_query': self.wf_query,
                'name': self.name,
                'launchdir': self.launchdir,
                'unique_launchdir': self.unique_launchdir,
                'sleep_time': self.sleep_time}

    @classmethod
    @recursive_deserialize
    def from_dict(cls, m_dict):
        assert getattr(cls, '_fw_name') == m_dict.pop('_fw_name')
        launchpad = LaunchPad.from_dict(m_dict.pop('launchpad'))
        return cls(launchpad=launchpad, **m_dict)

    def start(self, raise_exception=False):
        """Start a launcher thread"""
        if self.thread and self.thread.is_alive():
            self.logger.warning('launcher thread is already running')  # not covered
        elif fw_config.MONGOMOCK_SERVERSTORE_FILE is None:
            self.event = threading.Event()
            self.thread = threading.Thread(target=self._launcher, args=(self.event,))
            self.thread.start()
        else:
            msg = 'launcher thread cannot be used with Mongomock'
            exc = ConfigurationException(msg)
            if raise_exception:
                raise exc
            print(exc)

    def stop(self, join=False):
        """Gracefully stop the launcher thread if it is running"""
        if self.event and not self.event.is_set():
            self.event.set()
            self.logger.info('stopping the launcher thread')
            if join:
                self.thread.join()

    def _get_launcher_fw_ids(self):
        """update the list of fw_ids to process in the launcher"""
        fw_query = {'state': 'READY', 'spec._category': {'$in': ['batch', 'interactive']}}
        if self.nodes_torun is not None:
            if len(self.nodes_torun) == 0:
                return []
            fw_query['fw_id'] = {'$in': self.nodes_torun}
        return self.launchpad.get_fw_ids_in_wfs(self.wf_query, fw_query)

    def _launcher(self, stop_event):
        """
        The main loop of the launcher

        Args:
            stop_event (threading.Event): an object used to quit the launcher
        """
        while not stop_event.is_set():
            fw_ids = self._get_launcher_fw_ids()
            while fw_ids:
                if stop_event.is_set():
                    self.logger.info('launcher thread stopped')
                    return
                for fw_id in fw_ids:
                    fw_p = {'spec._category': True}
                    fw = self.launchpad.fireworks.find_one({'fw_id': fw_id}, fw_p)
                    try:
                        if fw['spec']['_category'] == 'batch':
                            self.qlaunch(fw_id)
                        else:
                            self.rlaunch(fw_id)
                    except Exception as err:  # very broad exception due to fireworks
                        self.logger.exception('Launch error: %s', str(err))
                fw_ids = self._get_launcher_fw_ids()
            sleep(self.sleep_time)
        self.logger.info('launcher thread stopped')

    def add_node(self, func, inputs, outputs=None, name=None, kwargs=None,
                 category=None, fworker=None, qadapter=None):
        """
        Add a python function node to an existing workflow

        Args:
            func (str): a function name with an optional module name in the
                format 'module.function'
            inputs ([tuple]): a list of positional arguments for the provided
                function. Every input is described by a tuple (fw_id, name, value)
                with the following elements:
                fw_id (int): The fw_id of a parent node providing the input; if the
                input is provided as a constant value, then None should be specified.
                name (str): The name of the input as provided in the list of outputs of
                the parent node;
                value: The value of the input; if output data from a parent node
                is used as input, then this should be set to None.

            outputs ([str]): names of the outputs
            name (str, None): name of the node
            kwargs (dict, None): a dictionary of keyword arguments for func
            category (str, None): job category, either 'batch' or 'interactive'
            fworker (FWorker, None): fworker for executing the batch jobs
            qadapter (CommonAdapter, None): qadapter for submitting batch jobs
        """
        if outputs is None:
            outputs = []
        if kwargs is None:
            kwargs = {}
        try:
            if fw_config.MONGOMOCK_SERVERSTORE_FILE is not None and category == 'batch':
                raise ConfigurationException('cannot add batch node with Mongomock')
            node_ids = list(set(i[0] for i in inputs if i[0] is not None))
            # check that all parent nodes are in the managed wfs
            if not all(n in self.fw_ids for n in node_ids):  # not covered
                raise ConfigurationException('some parent nodes are not in engine')
            # check that all parent nodes are in one worfklow
            qres = self.launchpad.get_wf_ids({'nodes': {'$in': node_ids}})
            if len(qres) <= 0:
                raise ConfigurationException('no valid parent nodes defined')
            if len(qres) != 1:
                raise ConfigurationException('some parent nodes are not in one workflow')
            # later, add all upstream fireworks of the nodes from other workflows
            # later check the outputs of parent nodes to match the inputs
            inps = [i[1] for i in inputs]
            # check that any equal inputs have the same source
            msg = 'input "{}" has more than one source'
            for inp in set(inps):
                sources = set(i[0] for i in inputs if i[1] == inp)
                if len(sources) != 1:
                    raise ConfigurationException(msg.format(inp))

            task = PyTask(func=func, inputs=inps, outputs=outputs, kwargs=kwargs)
            spec = {}
            spec['_dupefinder'] = {'_fw_name': 'DupeFinderExact'}
            spec['_fworker'] = fworker or self.name
            spec['_category'] = category or 'interactive'
            if spec['_category'] == 'batch':
                spec['_qadapter'] = qadapter
            locs = {i[1]: i[2] for i in inputs if i[0] is None}
            spec.update(locs)
            wflow = Workflow([Firework(tasks=[task], spec=spec, name=name)])
            self.launchpad.append_wf(wflow, fw_ids=node_ids)
        except Exception as exception:
            self.logger.error(exception, exc_info=1)
            raise

    def add_workflow(self, workflow=None, fw_id=None):
        """
        Add a workflow to the engine.
        Either a workflow object or a fw_id must be defined.

        Args:
            workflow (Workflow, None): a workflow object
            fw_id (int, None): a fw_id of a workflow existing on the launchpad
        """
        try:
            if ((workflow is None and fw_id is None)
                    or (workflow is not None and fw_id is not None)):
                msg = 'Either a workflow object or fw_id is needed.'
                raise ConfigurationException(msg)
            if fw_id is None:
                if not isinstance(workflow, (Workflow, Firework)):
                    msg = 'workflow must be instance of Workflow or Firework'
                    raise ConfigurationException(msg)
                fw_id = list(self.launchpad.add_wf(workflow).values())[0]
            else:
                if not isinstance(fw_id, int):
                    raise ConfigurationException('fw_id must be integer')
                if fw_id in self.fw_ids:
                    raise ConfigurationException('workflow already in engine')
                fw_ids = self.launchpad.get_fw_ids({'fw_id': fw_id})
                if len(fw_ids) != 1:
                    raise ConfigurationException('no workflow with this ID')
            self.append_wf_id(fw_id)
            self.logger.info('Added workflow ID: %d', fw_id)
        except Exception as exception:
            self.logger.error(exception, exc_info=1)
            raise

    def remove_workflow(self, fw_id):
        """
        Remove a workflow from the engine (but not deleted from launchpad)

        Args:
            fw_id (int): a fw_id of a node in the workflow to remove
        """
        self.remove_wf_id(fw_id)
        self.logger.info('Removed workflow ID: %d', fw_id)

    def update_node(self, fw_id, update_dict):
        """
        Update (modify) a workflow node.
        Only nodes in FIZZLED, DEFUSED, PAUSED, WAITING, and READY states
        can be modified.

        Args:
           fw_id (int): the fw_id of the node to modify
           update_dict (dict): a dictionary with the updates to perform
        """
        allowed_states = ['FIZZLED', 'DEFUSED', 'PAUSED', 'WAITING', 'READY']
        try:
            if fw_id not in self.fw_ids:
                msg = f'Node ID not configured in engine: {fw_id}'
                raise ConfigurationException(msg)
            fw = self.launchpad.fireworks.find_one({'fw_id': fw_id}, {'state': True})
            if fw['state'] not in allowed_states:
                msg = f'Node state must be in {allowed_states}'
                raise InvalidStateException(msg, fw_id)
            self.launchpad.update_spec([fw_id], update_dict)
        except Exception as exception:
            self.logger.error(exception, exc_info=1)
            raise

    def rerun_node(self, fw_id):
        """
        Rerun a workflow node.
        Nodes in FIZZLED and COMPLETED states will be directly rerun.
        Nodes in PAUSED and DEFUSED states are resumed and reignited, respectively.
        In all cases the target node state is WAITING. The final state can be
        READY if all parents are COMPLETED.

        Args:
            fw_id: the fw_id of the node to rerun
        """
        allowed_states = ['FIZZLED', 'DEFUSED', 'PAUSED', 'COMPLETED']
        try:
            if fw_id not in self.fw_ids:
                msg = f'Node ID not configured in engine: {fw_id}'
                raise ConfigurationException(msg)
            fw = self.launchpad.fireworks.find_one({'fw_id': fw_id}, {'state': True})
            if fw['state'] not in allowed_states:
                msg = f'Node state must be in {allowed_states}'
                raise InvalidStateException(msg, fw_id)
            if fw['state'] in ['FIZZLED', 'COMPLETED']:
                self.launchpad.rerun_fw(fw_id)
            elif fw['state'] == 'DEFUSED':
                self.launchpad.reignite_fw(fw_id)
            else:
                self.launchpad.resume_fw(fw_id)
        except Exception as exception:
            self.logger.error(exception, exc_info=1)
            raise

    def update_rerun_node(self, fw_id, update_dict):
        """
        Update (modify) and rerun a workflow node combined in one function.
        Only nodes in FIZZLED, DEFUSED, PAUSED, and COMPLETED states can be
        processed.

        Args:
            fw_id (int): the fw_id of the node to process
            update_dict (dict): a dictionary with the updates to perform
        """
        allowed_states = ['FIZZLED', 'DEFUSED', 'PAUSED', 'COMPLETED']
        try:
            if fw_id not in self.fw_ids:
                msg = f'Node ID not configured in engine: {fw_id}'
                raise ConfigurationException(msg)
            fw = self.launchpad.fireworks.find_one({'fw_id': fw_id}, {'state': True})
            if fw['state'] not in allowed_states:
                msg = f'Node state must be in {allowed_states}'
                raise InvalidStateException(msg, fw_id)
            if fw['state'] == 'COMPLETED':
                self.launchpad.defuse_fw(fw_id, rerun_duplicates=False)
                self.launchpad.update_spec([fw_id], update_dict)
                self.launchpad.reignite_fw(fw_id)
            elif fw['state'] == 'FIZZLED':
                self.launchpad.update_spec([fw_id], update_dict)
                self.launchpad.rerun_fw(fw_id, rerun_duplicates=False)
            elif fw['state'] == 'DEFUSED':
                self.launchpad.update_spec([fw_id], update_dict)
                self.launchpad.reignite_fw(fw_id)
            else:
                self.launchpad.update_spec([fw_id], update_dict)
                self.launchpad.resume_fw(fw_id)
        except Exception as exception:
            self.logger.error(exception, exc_info=1)
            raise

    def cancel_job(self, fw_id, restart=False, deactivate=False):
        """
        Cancel the execution of a node in RESERVED or RUNNING state.
        Either restart or deactivate can be set to True if required.

        Args:
            fw_id (int): the fw_id of the node to cancel
            restart (bool): restart node after cancelling
            deactivate (bool): deactivate node after cancelling
        """
        allowed_states = ['RESERVED', 'RUNNING']
        try:
            if fw_id not in self.fw_ids:
                msg = f'Node ID not configured in engine: {fw_id}'
                raise ConfigurationException(msg)
            if restart == deactivate:
                msg = 'Either restart or deactivate must be set to True'
                raise ConfigurationException(msg)
            fw_p = {'state': True, 'spec._category': True}
            fw = self.launchpad.fireworks.find_one({'fw_id': fw_id}, fw_p)
            if fw['state'] not in allowed_states:
                msg = f'Node state must be in {allowed_states}'
                raise InvalidStateException(msg, fw_id)
            reserve_id = self.launchpad.get_reservation_id_from_fw_id(fw_id)
            if fw['spec']['_category'] == 'batch':
                exec_cancel(reserve_id)
            if fw['state'] == 'RESERVED':  # not covered
                self.launchpad.cancel_reservation_by_reservation_id(reserve_id)
                if deactivate:
                    self.launchpad.pause_fw(fw_id)
            elif fw['state'] == 'RUNNING':  # not covered
                if restart:
                    self.launchpad.rerun_fw(fw_id)
                else:
                    self.launchpad.defuse_fw(fw_id)
        except Exception as exception:
            self.logger.error(exception, exc_info=1)
            raise



"""create a resource configuration"""
import os
import warnings
from argparse import ArgumentParser
from virtmat.middleware.resconfig import get_resconfig_loc, configure, setup_resconfig
from virtmat.middleware.utilities import format_warning_wrapper


def parse_clargs():
    """parse the command line arguments for the script"""
    m_description = 'create a resource configuration file from scratch'
    parser = ArgumentParser(description=m_description)
    add_arguments(parser)
    return parser.parse_args()


def add_arguments(parser):
    """add arguments to a parser object"""
    parser.add_argument('-i', '--interactive', default=False, required=False,
                        action='store_true', help='enable interactive mode, no overwrite')
    parser.add_argument('-o', '--overwrite', default=False, required=False,
                        action='store_true', help='overwrite existing configuration')
    parser.add_argument('-f', '--path', type=str, default=None, required=False,
                        help='non-default resconfig location, overrides RESCONFIG_LOC')


def resconfig_main(clargs):
    """main resconfig function"""
    warnings.formatwarning = format_warning_wrapper(warnings.formatwarning)
    resconfig_loc = clargs.path or get_resconfig_loc()
    if clargs.interactive:
        setup_resconfig(resconfig_loc)
        return
    if os.path.exists(resconfig_loc) and not clargs.overwrite:
        print(f'Resource configuration {resconfig_loc} exists.')
        return
    configure(resconfig_loc)


def main():
    """main function in module used as entry point"""
    resconfig_main(parse_clargs())
